import argparse
import logging

import yaml
import zmq


def manual_run(request, cfg):
    """
    Run the report service manually from any machine
    and provide the requested configuration for
    reports generation.

    :param request: a dictionary for generating reports for the
                     requested Instruments. This dict consists
                     of two keys. A upload boolean key for defining
                     the fate of the generated plots
                     (pushed to the DC repo. or staying locally)
                     and a key request which can either be a list
                     of requested Instruments names e.g ['SPB'] or ['all']
                     for generating reports, or dict or new requested
                     configuration.

    """


    port = cfg['GLOBAL']['server-port']
    con = zmq.Context()
    socket = con.socket(zmq.REQ)
    socket.connect(port)
    socket.send_pyobj(request)
    msg = socket.recv_pyobj()
    logging.info('{} Manual Run'.format(msg))


arg_parser = argparse.ArgumentParser(description='Manual Launch')
arg_parser.add_argument('--instrument', default=['all'], nargs='+',
                        help='Select the requested instruments. '
                             'Default=\"all\", which can be used for selecting'
                             ' all instruments.')
arg_parser.add_argument('--config-file', type=str,
                        default='./report_conf.yaml',
                        help='path to report configuration file '
                             'Default=./report_conf.yaml')
arg_parser.add_argument('--upload', action='store_true',
                        help='Required for uploading the generated figures.'
                             'Default=False. '
                             'Note: THIS HAS NO EFFECT IN SIM MODE!')
arg_parser.set_defaults(upload=False)
arg_parser.add_argument('--overwrite-conf', action='store_true',
                        help='A flag for using a different config file than'
                             'what is used by the running report_service.'
                             'Default=False, type=str.')
arg_parser.set_defaults(overwrite_conf=False)
arg_parser.add_argument('--report-fmt', default='html',
                        type=str, choices=['pdf', 'html'],
                        help='If available in the report service running mode,'
                             ' this can configure the report format '
                             'to be html or pdf. Default=html '
                             'Note: THIS HAS NO EFFECT IN PROD AND SIM MODES!')
arg_parser.add_argument('--log-file', type=str, default='./report.log',
                        help='The report log file path. Default=./report.log')
arg_parser.add_argument('--logging', type=str, default="INFO", 
                        help='logging modes: INFO, DEBUG or ERROR. '
                             'Default=INFO',
                        choices=['INFO', 'DEBUG', 'ERROR'])

if __name__ == "__main__":
    args = vars(arg_parser.parse_args())
    conf_file = args["config_file"]
    with open(conf_file, "r") as f:
        cfg = yaml.load(f.read(), Loader=yaml.FullLoader)

    logfile = args["log_file"]
    fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    logging.basicConfig(filename=logfile, filemode='a+',
                        level=getattr(logging, args['logging']),
                        format='%(levelname)-6s: %(asctime)s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    request = {'upload': args["upload"],
               'report-fmt': args["report_fmt"]}

    if args["overwrite_conf"]:
        request['req'] = cfg
    else:
        request['req'] = args["instrument"]
    manual_run(request, cfg)
